from bisect import bisect_left
from decimal import Decimal
from fractions import Fraction
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple
from warnings import warn

from eth_typing import ChecksumAddress
from eth_utils import to_checksum_address
from web3 import Web3

from ..config import get_web3
from ..exceptions import (
    DeprecationError,
    ExternalUpdateError,
    LiquidityPoolError,
    NoPoolStateAvailable,
    ZeroSwapError,
)
from ..logging import logger
from ..manager import AllPools, Erc20TokenHelperManager
from ..baseclasses import PoolHelper
from .abi import CAMELOT_POOL_ABI, UNISWAP_V2_POOL_ABI
from .v2_dataclasses import UniswapV2PoolSimulationResult, UniswapV2PoolState
from .v2_functions import generate_v2_pool_address

if TYPE_CHECKING:
    from ..erc20_token import Erc20Token

class LiquidityPool(PoolHelper):
    """
    Represents a Uniswap V2 liquidity pool
    """

    uniswap_version = 2

    __slots__: Tuple[str, ...] = (
        "_pool_state_archive",
        "_ratio_token0_in",
        "_ratio_token1_in",
        "_update_method",
        "_w3",
        "_w3_contract",
        "abi",
        "address",
        "factory",
        "fee_token0",
        "fee_token1",
        "fee",
        "name",
        "new_reserves",
        "reserves_token0",
        "reserves_token1",
        "router",
        "state",
        "token0_max_swap",
        "token0",
        "token1_max_swap",
        "token1",
        "update_block",
    )

    def __init__(
        self,
        address: str,
        tokens: Optional[List["Erc20Token"]] = None,
        name: Optional[str] = None,
        update_method: str = "polling",
        abi: Optional[list] = None,
        factory_address: Optional[str] = None,
        factory_init_hash: Optional[str] = None,
        # default fee for most UniswapV2 AMMs is 0.3%
        fee: Fraction = Fraction(3, 1000),
        fee_token0: Optional[Fraction] = None,
        fee_token1: Optional[Fraction] = None,
        silent: bool = False,
        update_reserves_on_start: bool = True,
        unload_brownie_contract_after_init: bool = False,  # deprecated
        state_block: Optional[int] = None,
    ) -> None:
        """
        Create a new `LiquidityPool` object for interaction with a Uniswap
        V2 pool.

        Arguments
        ---------
        address : str
            Address for the deployed pool contract.
        tokens : List[Erc20Token], optional
            "Erc20Token" objects for the tokens held by the deployed pool.
        name : str, optional
            Name of the contract, e.g. "DAI-WETH".
        update_method : str
            A string that sets the method used to fetch updates to the pool.
            Can be "polling", which fetches updates from the chain object
            using the contract object, or "external" which relies on updates
            being provided from outside the object.
        abi : list, optional
            Contract ABI.
        factory_address : str, optional
            The address for the factory contract. The default assumes a
            mainnet Uniswap V2 factory contract. If creating a
            `LiquidityPool` object based on another ecosystem, provide this
            value or the address check will fail.
        factory_init_hash : str, optional
            The init hash for the factory contract. The default assumes a
            mainnet Uniswap V2 factory contract.
        fee : Fraction
            The swap fee imposed by the pool. Defaults to `Fraction(3,1000)`
            which is equivalent to 0.3%.
        fee_token0 : Fraction, optional
            Swap fee for token0. Same purpose as `fee` except useful for
            pools with different fees for each token.
        fee_token1 : Fraction, optional
            Swap fee for token1. Same purpose as `fee` except useful for
            pools with different fees for each token.
        silent : bool
            Suppress status output.
        update_reserves_on_start : bool
            Update the reserves during instantiation.
        state_block: int, optional
            Fetch initial state values from the chain at a particular block
            height. Defaults to the latest block if omitted.
        """

        if unload_brownie_contract_after_init:
            warn(
                "unload_brownie_contract_after_init is no longer needed and is "
                "ignored. Remove constructor argument to stop seeing this "
                "message."
            )

        self.address: ChecksumAddress = to_checksum_address(address)
        self.abi = abi or UNISWAP_V2_POOL_ABI

        _web3 = get_web3()
        if _web3 is not None:
            self._w3 = _web3
        else:
            from brownie import web3 as brownie_web3  # type: ignore[import]

            if brownie_web3.isConnected():
                self._w3 = brownie_web3
            else:
                raise ValueError("No connected web3 object provided.")

        self._w3_contract = self._w3.eth.contract(
            address=self.address,
            abi=self.abi,
        )

        if isinstance(fee, Decimal):
            warn(
                f"WARNING: fee set as a Decimal value instead of Fraction. The fee has been converted inside the LP helper from {repr(fee)} to {repr(Fraction(fee))}, please adjust your code to specify a Fraction to remove this warning. e.g. Fraction(3,1000) is equivalent to Decimal('0.003')."
            )
            fee = Fraction(fee)

        if factory_address is None != factory_init_hash is None:
            raise ValueError(
                f"Init hash not provided for factory {factory_address}"
            )

        if not isinstance(fee, Fraction):
            raise TypeError(
                f"LP fee was not correctly passed! "
                f"Expected '{Fraction().__class__.__name__}', "
                f"was '{fee.__class__.__name__}'"
            )

        self.fee_token0 = fee_token0 if fee_token0 is not None else fee
        self.fee_token1 = fee_token1 if fee_token1 is not None else fee

        if self.fee_token0 and self.fee_token1:
            self.fee = None
        else:
            self.fee = fee

        self._update_method = update_method
        self._ratio_token0_in: Optional[Decimal] = None
        self._ratio_token1_in: Optional[Decimal] = None
        self.token0_max_swap = 0
        self.token1_max_swap = 0
        self.new_reserves = False
        self.update_block = (
            state_block if state_block else self._w3.eth.get_block_number()
        )

        self.factory = self._w3_contract.functions.factory().call()

        # if a token pair was provided, check and set pointers for token0 and token1
        if tokens is not None:
            if len(tokens) != 2:
                raise ValueError(f"Expected 2 tokens, found {len(tokens)}")
            for token in tokens:
                if (
                    token.address
                    == self._w3_contract.functions.token0().call()
                ):
                    self.token0 = token
                elif (
                    token.address
                    == self._w3_contract.functions.token1().call()
                ):
                    self.token1 = token
                else:
                    raise ValueError(f"{token} not found in pool {self}")
        else:
            _token_manager = Erc20TokenHelperManager(self._w3.eth.chain_id)
            self.token0 = _token_manager.get_erc20token(
                address=self._w3_contract.functions.token0().call(),
                silent=silent,
            )
            self.token1 = _token_manager.get_erc20token(
                address=self._w3_contract.functions.token1().call(),
                silent=silent,
            )

        if factory_address is not None and factory_init_hash is not None:
            computed_pool_address = generate_v2_pool_address(
                token_addresses=[self.token0.address, self.token1.address],
                factory_address=factory_address,
                init_hash=factory_init_hash,
            )
            if computed_pool_address != self.address:
                raise ValueError(
                    f"Pool address {self.address} does not match deterministic address {computed_pool_address} from factory"
                )

        if name is not None:
            self.name = name
        else:
            if self.fee is not None:
                fee_string = (
                    f"{100*self.fee.numerator/self.fee.denominator:.2f}"
                )
            elif self.fee_token0 is not None and self.fee_token1 is not None:
                if self.fee_token0 != self.fee_token1:
                    fee_string = f"{100*self.fee_token0.numerator/self.fee_token0.denominator:.2f}/{100*self.fee_token1.numerator/self.fee_token1.denominator:.2f}"
                elif self.fee_token0 == self.fee_token1:
                    fee_string = f"{100*self.fee_token0.numerator/self.fee_token0.denominator:.2f}"

            self.name = f"{self.token0}-{self.token1} (V2, {fee_string}%)"

        if update_reserves_on_start:
            (
                self.reserves_token0,
                self.reserves_token1,
                *_,
            ) = self._w3_contract.functions.getReserves().call(
                block_identifier=self.update_block
            )[
                :2
            ]
        else:
            self.reserves_token0 = self.reserves_token1 = 0

        if self._update_method == "event":
            raise ValueError(
                "The 'event' update method is inaccurate and unsupported, please update your bot to use the default 'polling' method"
            )

        self.state = UniswapV2PoolState(
            pool=self,
            reserves_token0=self.reserves_token0,
            reserves_token1=self.reserves_token1,
        )
        self._pool_state_archive: Dict[int, UniswapV2PoolState] = {
            self.update_block: self.state
        }  # WIP

        AllPools(self._w3.eth.chain_id)[self.address] = self

        if not silent:
            logger.info(self.name)
            logger.info(
                f"• Token 0: {self.token0} - Reserves: {self.reserves_token0}"
            )
            logger.info(
                f"• Token 1: {self.token1} - Reserves: {self.reserves_token1}"
            )

    def __eq__(self, other) -> bool:
        if issubclass(type(other), PoolHelper):
            return self.address == other.address
        elif isinstance(other, str):
            return self.address.lower() == other.lower()
        else:
            raise NotImplementedError

    # The Brownie contract object cannot be pickled, so remove it and return the state
    def __getstate__(self) -> dict:
        keys_to_remove = ("_w3", "_w3_contract")

        try:
            self.__slots__
        except AttributeError:
            pass
        else:
            return {
                attr_name: getattr(self, attr_name, None)
                for attr_name in self.__slots__
                if attr_name not in keys_to_remove
            }

        return {
            key: value
            for key, value in self.__dict__.items()
            if key not in keys_to_remove
        }

    def __hash__(self):
        return hash(self.address)

    def __repr__(self):
        return f"LiquidityPool(address={self.address}, token0={self.token0}, token1={self.token1})"

    def __setstate__(self, state: Dict):
        for key, value in state.items():
            setattr(self, key, value)

    def __str__(self):
        return self.name

    def _update_pool_state(self):
        self.state = UniswapV2PoolState(
            pool=self,
            reserves_token0=self.reserves_token0,
            reserves_token1=self.reserves_token1,
        )

    def calculate_tokens_in_from_ratio_out(self) -> None:
        """
        Calculates the maximum token inputs for the target output ratios at current pool reserves
        """

        # token0 in, token1 out
        # formula: dx = y0*C - x0/(1-FEE), where C = token0/token1
        if self._ratio_token0_in:
            self.token0_max_swap = max(
                0,
                int(self.reserves_token1 * self._ratio_token0_in)
                - int(self.reserves_token0 / (1 - self.fee_token0)),
            )
        else:
            self.token0_max_swap = 0

        # token1 in, token0 out
        # formula: dy = x0*C - y0(1/FEE), where C = token1/token0
        if self._ratio_token1_in:
            self.token1_max_swap = max(
                0,
                int(self.reserves_token0 * self._ratio_token1_in)
                - int(self.reserves_token1 / (1 - self.fee_token1)),
            )
        else:
            self.token1_max_swap = 0

    def calculate_tokens_in_from_tokens_out(
        self,
        token_out_quantity: int,
        token_in: Optional["Erc20Token"] = None,
        token_out: Optional["Erc20Token"] = None,
        override_reserves_token0: Optional[int] = None,
        override_reserves_token1: Optional[int] = None,
        override_state: Optional[UniswapV2PoolState] = None,
    ) -> int:
        """
        Calculates the required token INPUT of token_in for a target OUTPUT
        at current pool reserves. Uses the `self.token0` and `self.token1`
        references to determine which token is being swapped in.

        @dev This method accepts overrides in the form of individual tokens
        reserves or a single override dictionary. The override dictionary is
        used by other helpers and is the preferred method. The individual
        overrides are left here for backward compatibility with older scripts,
        and will be deprecated in the future.
        """

        if override_state:
            if override_reserves_token0 or override_reserves_token1:
                raise ValueError(
                    "Provide a single override via `override_state` or individual reserves."
                )
            override_reserves_token0 = override_state.reserves_token0
            override_reserves_token1 = override_state.reserves_token1

        if (override_reserves_token0 and not override_reserves_token1) or (
            not override_reserves_token0 and override_reserves_token1
        ):
            raise ValueError(
                "Must provide reserve override values for both tokens"
            )

        if override_reserves_token0 and override_reserves_token1:
            logger.debug("Reserve overrides applied:")
            logger.debug(f"token0: {override_reserves_token0}")
            logger.debug(f"token1: {override_reserves_token1}")

        if token_in is not None:
            if token_in not in [self.token0, self.token1]:
                raise ValueError(
                    f"Could not identify token_in: {token_in}! This pool holds: {self.token0} {self.token1}"
                )
            if token_in == self.token0:
                reserves_in = (
                    override_reserves_token0
                    if override_reserves_token0 is not None
                    else self.reserves_token0
                )
                fee = self.fee_token0
                reserves_out = (
                    override_reserves_token1
                    if override_reserves_token1 is not None
                    else self.reserves_token1
                )
            elif token_in == self.token1:
                reserves_in = (
                    override_reserves_token1
                    if override_reserves_token1 is not None
                    else self.reserves_token1
                )
                reserves_out = (
                    override_reserves_token0
                    if override_reserves_token0 is not None
                    else self.reserves_token0
                )
                fee = self.fee_token1
            else:
                raise ValueError("wtf happened here? (token_in)")
        elif token_out is not None:
            if token_out not in [self.token0, self.token1]:
                raise ValueError(
                    f"Could not identify token_out: {token_out}! This pool holds: {self.token0} {self.token1}"
                )
            if token_out == self.token1:
                reserves_in = (
                    override_reserves_token0
                    if override_reserves_token0 is not None
                    else self.reserves_token0
                )
                reserves_out = (
                    override_reserves_token1
                    if override_reserves_token1 is not None
                    else self.reserves_token1
                )
                fee = self.fee_token0
            elif token_out == self.token0:
                reserves_in = (
                    override_reserves_token1
                    if override_reserves_token1 is not None
                    else self.reserves_token1
                )
                reserves_out = (
                    override_reserves_token0
                    if override_reserves_token0 is not None
                    else self.reserves_token0
                )
                fee = self.fee_token1
            else:
                raise ValueError("wtf happened here? (token_in)")

        # last token becomes infinitely expensive, so largest possible swap out is reserves - 1
        if token_out_quantity > reserves_out - 1:
            raise LiquidityPoolError(
                f"Requested amount out ({token_out_quantity}) >= pool reserves ({reserves_out})"
            )

        numerator = reserves_in * token_out_quantity * fee.denominator
        denominator = (reserves_out - token_out_quantity) * (
            fee.denominator - fee.numerator
        )
        return numerator // denominator + 1

    def calculate_tokens_out_from_tokens_in(
        self,
        token_in: "Erc20Token",
        token_in_quantity: int,
        override_reserves_token0: Optional[int] = None,
        override_reserves_token1: Optional[int] = None,
        override_state: Optional[UniswapV2PoolState] = None,
    ) -> int:
        """
        Calculates the expected token OUTPUT for a target INPUT at current pool reserves.
        Uses the self.token0 and self.token1 pointers to determine which token is being swapped in
        """

        if token_in_quantity <= 0:
            raise ZeroSwapError("token_in_quantity must be positive")

        if override_state:
            if override_reserves_token0 or override_reserves_token1:
                raise ValueError(
                    "Provide a single override via `override_state` or individual reserves."
                )
            override_reserves_token0 = override_state.reserves_token0
            override_reserves_token1 = override_state.reserves_token1

        if (override_reserves_token0 and not override_reserves_token1) or (
            not override_reserves_token0 and override_reserves_token1
        ):
            raise ValueError(
                "Must provide reserve override values for both tokens"
            )

        if override_reserves_token0 and override_reserves_token1:
            logger.debug("Reserve overrides applied:")
            logger.debug(f"token0: {override_reserves_token0}")
            logger.debug(f"token1: {override_reserves_token1}")

        if token_in == self.token0:
            reserves_in = (
                override_reserves_token0
                if override_reserves_token0 is not None
                else self.reserves_token0
            )
            reserves_out = (
                override_reserves_token1
                if override_reserves_token1 is not None
                else self.reserves_token1
            )
            fee = self.fee_token0
        elif token_in == self.token1:
            reserves_in = (
                override_reserves_token1
                if override_reserves_token1 is not None
                else self.reserves_token1
            )
            reserves_out = (
                override_reserves_token0
                if override_reserves_token0 is not None
                else self.reserves_token0
            )
            fee = self.fee_token1
        else:
            raise ValueError(
                f"Could not identify token_in: {token_in}! Pool holds: {self.token0} {self.token1}"
            )

        amount_in_with_fee = token_in_quantity * (
            fee.denominator - fee.numerator
        )
        numerator = amount_in_with_fee * reserves_out
        denominator = reserves_in * fee.denominator + amount_in_with_fee

        return numerator // denominator

    def restore_state_before_block(
        self,
        block: int,
    ) -> None:
        """
        Restore the last pool state recorded prior to a target block.

        Use this method to maintain consistent state data following a chain
        re-organization.
        """

        # Find the index for the most recent pool state PRIOR to the requested
        # block number.
        #
        # e.g. Calling restore_state_before_block(block=104) for a pool with
        # states at blocks 100, 101, 102, 103, 104. `bisect_left()` returns
        # block_index=3, since block 104 is at index=4. The state held at
        # index=3 is for block 103.
        # block_index = self._pool_state_archive.bisect_left(block)

        known_blocks = list(self._pool_state_archive.keys())
        block_index = bisect_left(known_blocks, block)

        if block_index == 0:
            raise NoPoolStateAvailable(
                f"No pool state known prior to block {block}"
            )

        # The last known state already meets the criterion, so return early
        if block_index == len(known_blocks):
            return

        # Remove states at and after the specified block
        for block in known_blocks[block_index:]:
            del self._pool_state_archive[block]

        restored_block, restored_state = list(
            self._pool_state_archive.items()
        )[-1]

        # Set mutable values to match state
        self.reserves_token0 = restored_state.reserves_token0
        self.reserves_token1 = restored_state.reserves_token1
        self.state = restored_state
        self.update_block = restored_block

    def set_swap_target(
        self,
        token_in: "Erc20Token",
        token_in_qty: int,
        token_out: "Erc20Token",
        token_out_qty: int,
        silent: bool = False,
    ) -> None:
        if not (
            (token_in == self.token0 and token_out == self.token1)
            or (token_in == self.token1 and token_out == self.token0)
        ):
            raise ValueError(
                "Tokens must match the two tokens held by this pool!"
            )

        if not silent:
            logger.info(
                f"{token_in} -> {token_out} @ ({token_in_qty} {token_in} = {token_out_qty} {token_out})"
            )

        if token_in == self.token0:
            # calculate the ratio of token0/token1 for swap of token0 -> token1
            self._ratio_token0_in = Decimal(
                (token_in_qty * 10**token_in.decimals)
            ) / Decimal(token_out_qty * 10**token_out.decimals)

        if token_in == self.token1:
            # calculate the ratio of token1/token0 for swap of token1 -> token0
            self._ratio_token1_in = Decimal(
                (token_in_qty * 10**token_in.decimals)
            ) / Decimal(token_out_qty * 10**token_out.decimals)

        self.calculate_tokens_in_from_ratio_out()

    def simulate_add_liquidity(
        self,
        added_reserves_token0: int,
        added_reserves_token1: int,
        override_state: Optional[UniswapV2PoolState] = None,
    ) -> UniswapV2PoolSimulationResult:
        if override_state:
            logger.debug(f"State override: {override_state}")

        reserves_token0 = (
            override_state.reserves_token0
            if override_state
            else self.reserves_token0
        )

        reserves_token1 = (
            override_state.reserves_token1
            if override_state
            else self.reserves_token1
        )

        return UniswapV2PoolSimulationResult(
            amount0_delta=added_reserves_token0,
            amount1_delta=added_reserves_token1,
            current_state=self.state,
            future_state=UniswapV2PoolState(
                pool=self,
                reserves_token0=reserves_token0 + added_reserves_token0,
                reserves_token1=reserves_token1 + added_reserves_token1,
            ),
        )

    def simulate_remove_liquidity(
        self,
        removed_reserves_token0: int,
        removed_reserves_token1: int,
        override_state: Optional[UniswapV2PoolState] = None,
    ) -> UniswapV2PoolSimulationResult:
        if override_state:
            logger.debug(f"State override: {override_state}")

        reserves_token0 = (
            override_state.reserves_token0
            if override_state
            else self.reserves_token0
        )

        reserves_token1 = (
            override_state.reserves_token1
            if override_state
            else self.reserves_token1
        )

        return UniswapV2PoolSimulationResult(
            amount0_delta=-removed_reserves_token0,
            amount1_delta=-removed_reserves_token1,
            current_state=self.state,
            future_state=UniswapV2PoolState(
                pool=self,
                reserves_token0=reserves_token0 - removed_reserves_token0,
                reserves_token1=reserves_token1 - removed_reserves_token1,
            ),
        )

    def simulate_swap(
        self,
        token_in: Optional["Erc20Token"] = None,
        token_in_quantity: Optional[int] = None,
        token_out: Optional["Erc20Token"] = None,
        token_out_quantity: Optional[int] = None,
        override_state: Optional[UniswapV2PoolState] = None,
    ) -> UniswapV2PoolSimulationResult:
        """
        TODO
        """

        if token_in_quantity is None and token_out_quantity is None:
            raise ValueError("No quantity was provided")

        if token_in_quantity is not None and token_out_quantity is not None:
            raise ValueError(
                "Provide token_in_quantity or token_out_quantity, not both"
            )

        if token_in and token_out and token_in == token_out:
            raise ValueError("Both tokens are the same!")

        if override_state:
            logger.debug(f"State override: {override_state}")

        if token_in and token_in not in (self.token0, self.token1):
            raise ValueError(
                f"Token not found! token_in = {repr(token_in)}, pool holds {self.token0},{self.token1}"
            )
        if token_out and token_out not in (self.token0, self.token1):
            raise ValueError(
                f"Token not found! token_out = {repr(token_out)}, pool holds {self.token0},{self.token1}"
            )

        if token_in is not None and token_in == self.token0:
            token_out = self.token1
        elif token_in is not None and token_in == self.token1:
            token_out = self.token0

        if token_out is not None and token_out == self.token0:
            token_in = self.token1
        elif token_out is not None and token_out == self.token1:
            token_in = self.token0

        # bugfix: (changed check `token_in_quantity is not None`)
        # swaps with zero amounts (a stupid value, but valid) were falling through
        # both blocks and function was returning None
        if token_in_quantity is not None and token_in is not None:
            token_out_quantity = self.calculate_tokens_out_from_tokens_in(
                token_in=token_in,
                token_in_quantity=token_in_quantity,
                # TODO: consolidate into single override_state arg
                override_reserves_token0=override_state.reserves_token0
                if override_state
                else None,
                override_reserves_token1=override_state.reserves_token1
                if override_state
                else None,
            )

            token0_delta = (
                -token_out_quantity
                if token_in is self.token1
                else token_in_quantity
            )
            token1_delta = (
                -token_out_quantity
                if token_in is self.token0
                else token_in_quantity
            )

        # bugfix: (changed check `token_out_quantity is not None`)
        # swaps with zero amounts (a stupid value, but valid) were falling through
        # both blocks and function was returning None
        elif token_out_quantity is not None:
            token_in_quantity = self.calculate_tokens_in_from_tokens_out(
                token_in=token_in,
                token_out=token_out,
                token_out_quantity=token_out_quantity,
                # TODO: consolidate into single override_state arg
                override_reserves_token0=override_state.reserves_token0
                if override_state
                else None,
                override_reserves_token1=override_state.reserves_token1
                if override_state
                else None,
            )

            token0_delta = (
                token_in_quantity
                if token_in == self.token0
                else -token_out_quantity
            )
            token1_delta = (
                token_in_quantity
                if token_in == self.token1
                else -token_out_quantity
            )

        return UniswapV2PoolSimulationResult(
            amount0_delta=token0_delta,
            amount1_delta=token1_delta,
            current_state=self.state,
            future_state=UniswapV2PoolState(
                pool=self,
                reserves_token0=self.reserves_token0 + token0_delta,
                reserves_token1=self.reserves_token1 + token1_delta,
            ),
        )

    def update_reserves(
        self,
        silent: bool = False,
        print_reserves: bool = True,
        print_ratios: bool = True,
        external_token0_reserves: Optional[int] = None,
        external_token1_reserves: Optional[int] = None,
        override_update_method: Optional[str] = None,
        update_block: Optional[int] = None,
    ) -> bool:
        """
        Checks for updated reserve values when set to "polling", otherwise
        if set to "external" assumes that provided reserves are valid
        """

        success = False

        # get the chain height from Brownie if a specific update_block is not provided
        if update_block is None:
            update_block = self._w3.eth.get_block_number()

        # discard stale updates, but allow updating the same pool multiple times per block (necessary if sending sync events individually)
        if update_block < self.update_block:
            raise ExternalUpdateError(
                f"Current state recorded at block {self.update_block}, received update for stale block {update_block}"
            )
        else:
            self.update_block = update_block

        if (
            self._update_method == "polling"
            or override_update_method == "polling"
        ):
            try:
                (
                    reserves0,
                    reserves1,
                    *_,
                ) = self._w3_contract.functions.getReserves().call(
                    block_identifier=self.update_block
                )
                if (self.reserves_token0, self.reserves_token1) != (
                    reserves0,
                    reserves1,
                ):
                    self.reserves_token0, self.reserves_token1 = (
                        reserves0,
                        reserves1,
                    )
                    if not silent:
                        logger.info(f"[{self.name}]")
                        if print_reserves:
                            logger.info(
                                f"{self.token0}: {self.reserves_token0}"
                            )
                            logger.info(
                                f"{self.token1}: {self.reserves_token1}"
                            )
                        if print_ratios:
                            logger.info(
                                f"{self.token0}/{self.token1}: {(self.reserves_token0/10**self.token0.decimals) / (self.reserves_token1/10**self.token1.decimals)}"
                            )
                            logger.info(
                                f"{self.token1}/{self.token0}: {(self.reserves_token1/10**self.token1.decimals) / (self.reserves_token0/10**self.token0.decimals)}"
                            )

                    # recalculate possible swaps using the new reserves
                    self.calculate_tokens_in_from_ratio_out()
                    self._update_pool_state()
                    self._pool_state_archive[update_block] = self.state
                    success = True
                else:
                    success = False
            except Exception as e:
                print(
                    f"LiquidityPool: Exception in update_reserves (polling): {e}"
                )
        elif self._update_method == "external":
            if not (
                (
                    external_token0_reserves is not None
                    and external_token1_reserves is not None
                )
            ):
                raise ValueError(
                    "Called update_reserves without providing reserve values for both tokens!"
                )

            # skip follow-up processing if the LP object already has the latest reserves, or if no reserves were provided
            if (
                external_token0_reserves == self.reserves_token0
                and external_token1_reserves == self.reserves_token1
            ):
                self.new_reserves = False
                success = False
            else:
                self.reserves_token0 = external_token0_reserves
                self.reserves_token1 = external_token1_reserves
                self.new_reserves = True
                self._update_pool_state()
                self._pool_state_archive[update_block] = self.state

            if not silent:
                logger.info(f"[{self.name}]")
                if print_reserves:
                    logger.info(f"{self.token0}: {self.reserves_token0}")
                    logger.info(f"{self.token1}: {self.reserves_token1}")
                if print_ratios:
                    logger.info(
                        f"{self.token0}/{self.token1}: {self.reserves_token0 / self.reserves_token1}"
                    )
                    logger.info(
                        f"{self.token1}/{self.token0}: {self.reserves_token1 / self.reserves_token0}"
                    )
            self.calculate_tokens_in_from_ratio_out()
            success = True
        elif self._update_method == "event":
            raise DeprecationError(
                "The 'event' update method is deprecated. Please update your bot to use the default 'polling' method"
            )
        else:
            success = False

        return success


class CamelotLiquidityPool(LiquidityPool):
    FEE_DENOMINATOR = 100_000

    def _calculate_tokens_out_from_tokens_in_stable_swap(
        self,
        token_in: "Erc20Token",
        token_in_quantity: int,
        override_state: Optional[dict] = None,
    ) -> int:
        """
        Calculates the expected token OUTPUT for a target INPUT at current pool reserves.
        Uses the self.token0 and self.token1 pointers to determine which token is being swapped in
        """

        override_reserves_token0: Optional[int] = None
        override_reserves_token1: Optional[int] = None

        if override_state is not None:
            try:
                override_reserves_token0 = override_state["reserves_token0"]
            except KeyError:
                pass
            else:
                logger.debug(f"{override_reserves_token0=}")

            try:
                override_reserves_token1 = override_state["reserves_token1"]
            except KeyError:
                pass
            else:
                logger.debug(f"{override_reserves_token1=}")

        if token_in_quantity <= 0:
            raise ZeroSwapError("token_in_quantity must be positive")

        precision_multiplier_token0 = 10**self.token0.decimals
        precision_multiplier_token1 = 10**self.token1.decimals

        def _k(balance_0, balance_1) -> int:
            _x: int = balance_0 * 10**18 // precision_multiplier_token0
            _y: int = balance_1 * 10**18 // precision_multiplier_token1
            _a: int = _x * _y // 10**18
            _b: int = (_x * _x // 10**18) + (_y * _y // 10**18)
            return _a * _b // 10**18  # x^3*y+y^3*x >= k

        def _get_y(x_0: int, xy: int, y: int) -> int:
            for _ in range(255):
                y_prev = y
                k = _f(x_0, y)
                if k < xy:
                    dy = (xy - k) * 10**18 // _d(x_0, y)
                    y = y + dy
                else:
                    dy = (k - xy) * 10**18 // _d(x_0, y)
                    y = y - dy

                if y > y_prev:
                    if y - y_prev <= 1:
                        return y
                else:
                    if y_prev - y <= 1:
                        return y

            return y

        def _f(x_0: int, y: int) -> int:
            return (
                x_0 * (y * y // 10**18 * y // 10**18) // 10**18
                + (x_0 * x_0 // 10**18 * x_0 // 10**18) * y // 10**18
            )

        def _d(x_0: int, y: int) -> int:
            return 3 * x_0 * (y * y // 10**18) // 10**18 + (
                x_0 * x_0 // 10**18 * x_0 // 10**18
            )

        # fee_percent is stored as a uint16 in the contract, but as a Fraction
        # in this helper, so must be converted.
        #
        # e.g. 0.04% fee = Fraction(1,2500) in the helper, fee = 40 in the
        # contract. To convert, multiply the fraction by the `FEE_DENOMINATOR`,
        # so fee_percent = 1/2500 * 100000 = 40

        fee_percent = (
            self.fee_token0 if token_in is self.token0 else self.fee_token1
        ) * self.FEE_DENOMINATOR

        reserves_token0 = (
            override_reserves_token0
            if override_reserves_token0 is not None
            else self.reserves_token0
        )
        reserves_token1 = (
            override_reserves_token1
            if override_reserves_token1 is not None
            else self.reserves_token1
        )

        # remove fee from amount received
        token_in_quantity -= (
            token_in_quantity * fee_percent // self.FEE_DENOMINATOR
        )
        xy = _k(reserves_token0, reserves_token1)
        reserves_token0 = (
            reserves_token0 * 10**18 // precision_multiplier_token0
        )
        reserves_token1 = (
            reserves_token1 * 10**18 // precision_multiplier_token1
        )
        reserve_a, reserve_b = (
            (reserves_token0, reserves_token1)
            if token_in is self.token0
            else (reserves_token1, reserves_token0)
        )
        token_in_quantity = (
            token_in_quantity * 10**18 // precision_multiplier_token0
            if token_in is self.token0
            else token_in_quantity * 10**18 // precision_multiplier_token1
        )
        y = reserve_b - _get_y(token_in_quantity + reserve_a, xy, reserve_b)

        return (
            y
            * (
                precision_multiplier_token1
                if token_in is self.token0
                else precision_multiplier_token0
            )
            // 10**18
        )

    def __init__(
        self,
        address: str,
        tokens: Optional[List["Erc20Token"]] = None,
        name: Optional[str] = None,
        update_method: str = "polling",
        abi: Optional[list] = None,
        silent: bool = False,
        update_reserves_on_start: bool = True,
        unload_brownie_contract_after_init: bool = False,  # deprecated
    ) -> None:
        if unload_brownie_contract_after_init:
            warn(
                "unload_brownie_contract_after_init is no longer needed and is "
                "ignored. Remove constructor argument to stop seeing this "
                "message."
            )

        _web3 = get_web3()
        if _web3 is not None:
            _web3 = _web3
        else:
            from brownie import web3 as brownie_web3  # type: ignore[import]

            if brownie_web3.isConnected():
                _web3 = brownie_web3
            else:
                raise ValueError("No connected web3 object provided.")

        if TYPE_CHECKING:
            assert isinstance(_web3, Web3)

        address = to_checksum_address(address)

        _w3_contract = _web3.eth.contract(
            address=address, abi=abi or CAMELOT_POOL_ABI
        )

        stable_pool: bool = _w3_contract.functions.stableSwap().call()

        (
            _,
            _,
            fee_token0,
            fee_token1,
        ) = _w3_contract.functions.getReserves().call()
        fee_denominator = _w3_contract.functions.FEE_DENOMINATOR().call()
        fee_token0 = Fraction(fee_token0, fee_denominator)
        fee_token1 = Fraction(fee_token1, fee_denominator)

        super().__init__(
            address=address,
            tokens=tokens,
            name=name,
            update_method=update_method,
            abi=abi,
            fee_token0=fee_token0,
            fee_token1=fee_token1,
            silent=silent,
            update_reserves_on_start=update_reserves_on_start,
        )

        if stable_pool:
            # replace the calculate_tokens_out_from_tokens_in method for stable-only pools
            self.calculate_tokens_out_from_tokens_in = self._calculate_tokens_out_from_tokens_in_stable_swap  # type: ignore[assignment]



from ...constants import (
    MAX_INT128,
    MAX_INT256,
    MAX_UINT160,
    MIN_INT128,
    MIN_INT256,
)
from ...exceptions import EVMRevertError


def mulmod(x, y, k):
    if k == 0:
        raise EVMRevertError
    return (x * y) % k


# adapted from OpenZeppelin's overflow checks, which throw
# an exception if the input value exceeds the maximum value
# for this type
def to_int128(x):
    if not (MIN_INT128 <= x <= MAX_INT128):
        raise EVMRevertError(f"{x} outside range of int128 values")
    return x


def to_int256(x):
    if not (MIN_INT256 <= x <= MAX_INT256):
        raise EVMRevertError(f"{x} outside range of int256 values")
    return x


def to_uint160(x):
    if x > MAX_UINT160:
        raise EVMRevertError(f"{x} greater than maximum uint160 value")
    return x


# Dumb integer "conversions" that performs no value checking to mimic Solidity's
# inline typecasting for int/uint types. Makes copy-pasting the Solidity functions
# easier since in-line casts can remain
def int8(x):
    return x


def int16(x):
    return x


def int24(x):
    return x


def int32(x):
    return x


def int40(x):
    return x


def int48(x):
    return x


def int56(x):
    return x


def int64(x):
    return x


def int72(x):
    return x


def int80(x):
    return x


def int88(x):
    return x


def int96(x):
    return x


def int104(x):
    return x


def int112(x):
    return x


def int120(x):
    return x


def int128(x):
    return x


def int136(x):
    return x


def int144(x):
    return x


def int152(x):
    return x


def int160(x):
    return x


def int168(x):
    return x


def int176(x):
    return x


def int184(x):
    return x


def int192(x):
    return x


def int200(x):
    return x


def int208(x):
    return x


def int216(x):
    return x


def int224(x):
    return x


def int232(x):
    return x


def int240(x):
    return x


def int248(x):
    return x


def int256(x):
    return x


def uint8(x):
    return x


def uint16(x):
    return x


def uint24(x):
    return x


def uint32(x):
    return x


def uint40(x):
    return x


def uint48(x):
    return x


def uint56(x):
    return x


def uint64(x):
    return x


def uint72(x):
    return x


def uint80(x):
    return x


def uint88(x):
    return x


def uint96(x):
    return x


def uint104(x):
    return x


def uint112(x):
    return x


def uint120(x):
    return x


def uint128(x):
    return x


def uint136(x):
    return x


def uint144(x):
    return x


def uint152(x):
    return x


def uint160(x):
    return x


def uint168(x):
    return x


def uint176(x):
    return x


def uint184(x):
    return x


def uint192(x):
    return x


def uint200(x):
    return x


def uint208(x):
    return x


def uint216(x):
    return x


def uint224(x):
    return x


def uint232(x):
    return x


def uint240(x):
    return x


def uint248(x):
    return x


def uint256(x):
    return x



# Base exception
class DegenbotError(Exception):
    """
    Base exception, intended as a generic exception and a base class for
    for all more-specific exceptions raised by various degenbot modules
    """


class DeprecationError(ValueError):
    """
    Thrown when a feature, class, method, etc. is deprecated.

    Subclasses `ValueError` instead of `Exception`, less likely to be ignored.
    """


# 1st level exceptions (derived from `DegenbotError`)
class ArbitrageError(DegenbotError):
    """
    Exception raised inside arbitrage helpers
    """


class BlockUnavailableError(DegenbotError):
    """
    Exception raised when a call for a specific block fails (trie node unavailable)
    """


class Erc20TokenError(DegenbotError):
    """
    Exception raised inside ERC-20 token helpers
    """


class EVMRevertError(DegenbotError):
    """
    Thrown when a simulated EVM contract operation would revert
    """


class LiquidityPoolError(DegenbotError):
    """
    Exception raised inside liquidity pool helpers
    """


class ManagerError(DegenbotError):
    """
    Exception raised inside manager helpers
    """


class TransactionError(DegenbotError):
    """
    Exception raised inside transaction simulation helpers
    """


# 2nd level exceptions for Arbitrage classes
class ArbCalculationError(ArbitrageError):
    """
    Thrown when an arbitrage calculation fails
    """


class InvalidSwapPathError(ArbitrageError):
    """
    Thrown in arbitrage helper constructors when the provided path is invalid
    """

    pass


class ZeroLiquidityError(ArbitrageError):
    """
    Thrown by the arbitrage helper if a pool in the path has no liquidity in the direction of the proposed swap
    """


# 2nd level exceptions for Uniswap Liquidity Pool classes
class BitmapWordUnavailableError(LiquidityPoolError):
    """
    Thrown by the ported V3 swap function when the bitmap word is not available.
    This should be caught by the helper to perform automatic fetching, and should
    not be raised to the calling function
    """


class ExternalUpdateError(LiquidityPoolError):
    """
    Thrown when an external update does not pass sanity checks
    """


class MissingTickWordError(LiquidityPoolError):
    """
    Thrown by the TickBitmap library when calling for an operation on a word that
    should be available, but is not
    """


class NoPoolStateAvailable(LiquidityPoolError):
    """
    Thrown by the `restore_state_before_block` method when a previous pool
    state is not available. This can occur, e.g. if a pool was created in a
    block at or after a re-organization.
    """


class ZeroSwapError(LiquidityPoolError):
    """
    Thrown if a swap calculation resulted or would result in zero output
    """


# 2nd level exceptions for Transaction classes
class LedgerError(TransactionError):
    """
    Thrown when the ledger does not align with the expected state
    """


class TransactionEncodingError(TransactionError):
    """
    Thrown when a transaction input cannot be decoded using the known ABI
    """


# 2nd level exceptions for Uniswap Manager classes
class PoolNotAssociated(ManagerError):
    """
    Thrown by a UniswapV2LiquidityPoolManager or UniswapV3LiquidityPoolManager
    class if a requested pool address is not associated with the DEX.
    """



from typing import Dict
from eth_utils import to_checksum_address
from eth_typing import ChecksumAddress

MIN_INT16: int = -(2 ** (16 - 1))
MAX_INT16: int = (2 ** (16 - 1)) - 1

MIN_INT128: int = -(2 ** (128 - 1))
MAX_INT128: int = (2 ** (128 - 1)) - 1

MIN_INT256: int = -(2 ** (256 - 1))
MAX_INT256: int = (2 ** (256 - 1)) - 1

MIN_UINT8: int = 0
MAX_UINT8: int = 2**8 - 1

MIN_UINT128: int = 0
MAX_UINT128: int = 2**128 - 1

MIN_UINT160: int = 0
MAX_UINT160: int = 2**160 - 1

MIN_UINT256: int = 0
MAX_UINT256: int = 2**256 - 1

ZERO_ADDRESS: ChecksumAddress = to_checksum_address(
    "0x0000000000000000000000000000000000000000"
)


# Contract addresses for the native blockchain token, keyed by chain ID
WRAPPED_NATIVE_TOKENS: Dict[int, ChecksumAddress] = {
    # Ethereum (WETH)
    1: to_checksum_address("0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"),
    # Fantom (WFTM)
    250: to_checksum_address("0x21be370D5312f44cB42ce377BC9b8a0cEF1A4C83"),
    # Arbitrum (AETH)
    42161: to_checksum_address("0x82aF49447D8a07e3bd95BD0d56f35241523fBab1"),
    # Avalanche (WAVAX)
    43114: to_checksum_address("0xB31f66AA3C1e785363F0875A1B74E27b85FD66c7"),
}



from typing import Tuple

from . import full_math as FullMath
from . import sqrt_price_math as SqrtPriceMath
from .functions import uint256


def computeSwapStep(
    sqrtRatioCurrentX96: int,
    sqrtRatioTargetX96: int,
    liquidity: int,
    amountRemaining: int,
    feePips: int,
) -> Tuple[int, int, int, int]:
    zeroForOne: bool = sqrtRatioCurrentX96 >= sqrtRatioTargetX96
    exactIn: bool = amountRemaining >= 0

    if exactIn:
        amountRemainingLessFee: int = FullMath.mulDiv(
            uint256(amountRemaining), 10**6 - feePips, 10**6
        )
        amountIn = (
            SqrtPriceMath.getAmount0Delta(
                sqrtRatioTargetX96, sqrtRatioCurrentX96, liquidity, True
            )
            if zeroForOne
            else SqrtPriceMath.getAmount1Delta(
                sqrtRatioCurrentX96, sqrtRatioTargetX96, liquidity, True
            )
        )
        if amountRemainingLessFee >= amountIn:
            sqrtRatioNextX96 = sqrtRatioTargetX96
        else:
            sqrtRatioNextX96 = SqrtPriceMath.getNextSqrtPriceFromInput(
                sqrtRatioCurrentX96,
                liquidity,
                amountRemainingLessFee,
                zeroForOne,
            )
    else:
        amountOut = (
            SqrtPriceMath.getAmount1Delta(
                sqrtRatioTargetX96, sqrtRatioCurrentX96, liquidity, False
            )
            if zeroForOne
            else SqrtPriceMath.getAmount0Delta(
                sqrtRatioCurrentX96, sqrtRatioTargetX96, liquidity, False
            )
        )
        if uint256(-amountRemaining) >= amountOut:
            sqrtRatioNextX96 = sqrtRatioTargetX96
        else:
            sqrtRatioNextX96 = SqrtPriceMath.getNextSqrtPriceFromOutput(
                sqrtRatioCurrentX96,
                liquidity,
                uint256(-amountRemaining),
                zeroForOne,
            )

    max: bool = sqrtRatioTargetX96 == sqrtRatioNextX96
    # get the input/output amounts
    if zeroForOne:
        amountIn = (
            amountIn
            if (max and exactIn)
            else SqrtPriceMath.getAmount0Delta(
                sqrtRatioNextX96, sqrtRatioCurrentX96, liquidity, True
            )
        )
        amountOut = (
            amountOut
            if (max and not exactIn)
            else SqrtPriceMath.getAmount1Delta(
                sqrtRatioNextX96, sqrtRatioCurrentX96, liquidity, False
            )
        )
    else:
        amountIn = (
            amountIn
            if (max and exactIn)
            else SqrtPriceMath.getAmount1Delta(
                sqrtRatioCurrentX96, sqrtRatioNextX96, liquidity, True
            )
        )
        amountOut = (
            amountOut
            if (max and not exactIn)
            else SqrtPriceMath.getAmount0Delta(
                sqrtRatioCurrentX96, sqrtRatioNextX96, liquidity, False
            )
        )

    # cap the output amount to not exceed the remaining output amount
    if not exactIn and (amountOut > uint256(-amountRemaining)):
        amountOut = uint256(-amountRemaining)

    if exactIn and (sqrtRatioNextX96 != sqrtRatioTargetX96):
        # we didn't reach the target, so take the remainder of the maximum input as fee
        feeAmount = uint256(amountRemaining) - amountIn
    else:
        feeAmount = FullMath.mulDivRoundingUp(
            amountIn, feePips, 10**6 - feePips
        )

    return (
        sqrtRatioNextX96,
        amountIn,
        amountOut,
        feeAmount,
    )



from threading import Lock
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from web3 import Web3

from ..arbitrage.uniswap_lp_cycle import UniswapLpCycle
from ..constants import WRAPPED_NATIVE_TOKENS
from ..exceptions import ManagerError
from ..erc20_token import Erc20Token
from ..baseclasses import HelperManager
from ..uniswap.managers import (
    UniswapV2LiquidityPoolManager,
    UniswapV3LiquidityPoolManager,
)
from token_manager import Erc20TokenHelperManager

if TYPE_CHECKING:
    from ..baseclasses import ArbitrageHelper
    from ..uniswap.v2_liquidity_pool import LiquidityPool
    from ..uniswap.v3_liquidity_pool import V3LiquidityPool


class ArbitrageHelperManager(HelperManager):
    """
    A class that generates and tracks Arbitrage helpers

    The dictionary of arbitrage helpers is held as a class attribute, so all manager
    objects reference the same state data
    """

    _state: Dict = {}

    def __init__(self, chain_id: int):
        # the internal state data for this object is held in the
        # class-level _state dictionary, keyed by the chain ID
        if self._state.get(chain_id):
            self.__dict__ = self._state[chain_id]
        else:
            self._state[chain_id] = {}
            self.__dict__ = self._state[chain_id]

            # initialize internal attributes
            self._arbs: Dict = {}  # all known arbs, keyed by id
            self._blacklisted_ids: set = set()
            self._chain_id = chain_id
            self._erc20tokenmanager = Erc20TokenHelperManager(chain_id)
            self._lock = Lock()
            self._v2_pool_managers: Dict[
                str, UniswapV2LiquidityPoolManager
            ] = {}  # all V2 pool managers, keyed by factory address
            self._v3_pool_managers: Dict[
                str, UniswapV3LiquidityPoolManager
            ] = {}  # all V3 pool managers, keyed by factory address

    def add_pool_manager(self, factory_address: str, uniswap_version: int):
        """
        Create a Uniswap pool manager from the factory contract address and version, store in the internal dictionary of pool managers
        """

        if (
            factory_address in self._v2_pool_managers
            or factory_address in self._v3_pool_managers
        ):
            raise ValueError(
                f"Pool manager for factory={factory_address} already exists"
            )

        if uniswap_version == 2:
            self._v2_pool_managers[
                factory_address
            ] = UniswapV2LiquidityPoolManager(factory_address)
        elif uniswap_version == 3:
            self._v3_pool_managers[
                factory_address
            ] = UniswapV3LiquidityPoolManager(factory_address)
        else:
            raise ValueError

    def build(
        self,
        arb_type: str,
        swap_pools: Union[
            List[Union[LiquidityPool, V3LiquidityPool]],
            List[str],
        ],
        update_method: str = "polling",
        input_token: Optional[Union[str, Erc20Token]] = None,
    ) -> ArbitrageHelper:
        """
        Returns the arb helper
        """

        native_wrapped_token_address = WRAPPED_NATIVE_TOKENS[self._chain_id]

        print(native_wrapped_token_address)

        input_token = self._erc20tokenmanager.get_erc20token(
            native_wrapped_token_address
        )

        _swap_pools: List[Union[LiquidityPool, V3LiquidityPool]] = []

        # replace all pool addresses with helper objects
        for i, pool in enumerate(swap_pools):
            if isinstance(pool, str):
                # if an address was provided, get the pool helper object
                pool_helper: Union[LiquidityPool, V3LiquidityPool]

                # iterate through the pool managers (may be multiple compatible DEX on one chain)
                for v2_pool_manager in self._v2_pool_managers.values():
                    try:
                        pool_helper = v2_pool_manager.get_pool(pool)
                    except Exception:
                        pass
                for v3_pool_manager in self._v3_pool_managers.values():
                    try:
                        pool_helper = v3_pool_manager.get_pool(pool)
                    except Exception:
                        pass
                try:
                    pool_helper
                except Exception:
                    # will throw if the pool helper could not be found
                    raise ValueError(
                        f"Could not generate Uniswap LP helper for pool {pool}"
                    )
                else:
                    print(pool_helper)
            elif isinstance(pool, (LiquidityPool, V3LiquidityPool)):
                # otherwise, use the helper directly
                pool_helper = pool
            else:
                raise TypeError(
                    f"Pool {pool} is {type(pool)}! Expected LiquidityPool, V3LiquidityPool, or string"
                )

            _swap_pools[i] = pool_helper

        arb_id = Web3.keccak(
            hexstr="".join([pool.address[2:] for pool in _swap_pools])
        ).hex()

        # check if the helper is already known, throw exception if so
        try:
            arb_helper = self._arbs[arb_id]
        except KeyError:
            pass
        else:
            raise ValueError("Arbitrage helper already exists")

        arb_helper = UniswapLpCycle(
            input_token=input_token,
            swap_pools=_swap_pools,
            max_input=None,
            id=arb_id,
        )
        return arb_helper

    def get(
        self,
        arb_id: str,
        arb_type: str,
        chain_id: int,
        input_token: Optional[Union[str, Erc20Token]] = None,
        update_method: str = "polling",
    ) -> ArbitrageHelper:
        """
        Get an arbitrage path object from its ID and the type. An ID is a keccak address of all pool addresses, in order.

        Type can only be "cycle", but additional types will be added later
        """

        # attempt to retrieve the arb (might already exist)
        try:
            arb_helper = self._arbs[arb_id][arb_type]
        except KeyError:
            pass
        else:
            return arb_helper

        # otherwise create a new arb helper
        try:
            if arb_type == "cycle":
                if input_token is None:
                    native_wrapped_token_address = WRAPPED_NATIVE_TOKENS[
                        chain_id
                    ]
                    input_token = self._erc20tokenmanager.get_erc20token(
                        native_wrapped_token_address
                    )
                elif not isinstance(input_token, "Erc20Token"):
                    input_token = self._erc20tokenmanager.get_erc20token(
                        input_token
                    )
                # arb_helper = UniswapLpCycle(input_token=input_token)
                arb_helper = self.build(
                    arb_type=arb_type,
                    # TODO: read pools from a file,
                    # swap_pools=XXX,
                    update_method=update_method,
                    input_token=input_token,
                )
        except Exception:
            raise ManagerError(f"Could not create Arbitrage helper: {arb_id=}")
        else:
            return arb_helper



from decimal import Decimal

from ...constants import MAX_UINT128
from ...exceptions import EVMRevertError
from . import tick_math as TickMath
from .functions import uint24

# type hinting aliases
Int24 = int


def tickSpacingToMaxLiquidityPerTick(tickSpacing: Int24):
    if not (-(2**23) <= tickSpacing <= 2**23 - 1):
        raise EVMRevertError("input not a valid int24")

    minTick = Decimal(TickMath.MIN_TICK) // tickSpacing * tickSpacing
    maxTick = Decimal(TickMath.MAX_TICK) // tickSpacing * tickSpacing
    numTicks = uint24((maxTick - minTick) // tickSpacing) + 1

    result = MAX_UINT128 // numTicks

    if not (0 <= result <= MAX_UINT128):
        raise EVMRevertError("result not a valid uint128")

    return result



from ..baseclasses import ArbitrageHelper


class LpSwapWithFuture(ArbitrageHelper):
    def __init__(self, *args, **kwargs):
        raise DeprecationWarning(
            "This class has been deprecated, please transition to UniswapLpCycle."
        )



from ...constants import MAX_UINT256
from ...exceptions import EVMRevertError
from . import yul_operations as yul
from .functions import int24, int256, uint160, uint256

MIN_TICK = -887272
MAX_TICK = -MIN_TICK
MIN_SQRT_RATIO = 4295128739
MAX_SQRT_RATIO = 1461446703485210103287273052203988822378723970342


def getSqrtRatioAtTick(tick: int) -> int:
    abs_tick = uint256(-int256(tick)) if tick < 0 else uint256(int256(tick))
    if not (0 <= abs_tick <= uint256(MAX_TICK)):
        raise EVMRevertError("T")

    ratio = (
        0xFFFCB933BD6FAD37AA2D162D1A594001
        if (abs_tick & 0x1 != 0)
        else 0x100000000000000000000000000000000
    )

    if abs_tick & 0x2 != 0:
        ratio = (ratio * 0xFFF97272373D413259A46990580E213A) >> 128
    if abs_tick & 0x4 != 0:
        ratio = (ratio * 0xFFF2E50F5F656932EF12357CF3C7FDCC) >> 128
    if abs_tick & 0x8 != 0:
        ratio = (ratio * 0xFFE5CACA7E10E4E61C3624EAA0941CD0) >> 128
    if abs_tick & 0x10 != 0:
        ratio = (ratio * 0xFFCB9843D60F6159C9DB58835C926644) >> 128
    if abs_tick & 0x20 != 0:
        ratio = (ratio * 0xFF973B41FA98C081472E6896DFB254C0) >> 128
    if abs_tick & 0x40 != 0:
        ratio = (ratio * 0xFF2EA16466C96A3843EC78B326B52861) >> 128
    if abs_tick & 0x80 != 0:
        ratio = (ratio * 0xFE5DEE046A99A2A811C461F1969C3053) >> 128
    if abs_tick & 0x100 != 0:
        ratio = (ratio * 0xFCBE86C7900A88AEDCFFC83B479AA3A4) >> 128
    if abs_tick & 0x200 != 0:
        ratio = (ratio * 0xF987A7253AC413176F2B074CF7815E54) >> 128
    if abs_tick & 0x400 != 0:
        ratio = (ratio * 0xF3392B0822B70005940C7A398E4B70F3) >> 128
    if abs_tick & 0x800 != 0:
        ratio = (ratio * 0xE7159475A2C29B7443B29C7FA6E889D9) >> 128
    if abs_tick & 0x1000 != 0:
        ratio = (ratio * 0xD097F3BDFD2022B8845AD8F792AA5825) >> 128
    if abs_tick & 0x2000 != 0:
        ratio = (ratio * 0xA9F746462D870FDF8A65DC1F90E061E5) >> 128
    if abs_tick & 0x4000 != 0:
        ratio = (ratio * 0x70D869A156D2A1B890BB3DF62BAF32F7) >> 128
    if abs_tick & 0x8000 != 0:
        ratio = (ratio * 0x31BE135F97D08FD981231505542FCFA6) >> 128
    if abs_tick & 0x10000 != 0:
        ratio = (ratio * 0x9AA508B5B7A84E1C677DE54F3E99BC9) >> 128
    if abs_tick & 0x20000 != 0:
        ratio = (ratio * 0x5D6AF8DEDB81196699C329225EE604) >> 128
    if abs_tick & 0x40000 != 0:
        ratio = (ratio * 0x2216E584F5FA1EA926041BEDFE98) >> 128
    if abs_tick & 0x80000 != 0:
        ratio = (ratio * 0x48A170391F7DC42444E8FA2) >> 128

    if tick > 0:
        ratio = (MAX_UINT256) // ratio

    # this divides by 1<<32 rounding up to go from a Q128.128 to a Q128.96
    # we then downcast because we know the result always fits within 160 bits due to our tick input constraint
    # we round up in the division so getTickAtSqrtRatio of the output price is always consistent
    return uint160((ratio >> 32) + (0 if (ratio % (1 << 32) == 0) else 1))


def getTickAtSqrtRatio(sqrt_price_x96: int) -> int:
    if not (0 <= sqrt_price_x96 <= 2**160 - 1):
        raise EVMRevertError("not a valid uint160")

    # second inequality must be < because the price can never reach the price at the max tick
    if not (
        sqrt_price_x96 >= MIN_SQRT_RATIO and sqrt_price_x96 < MAX_SQRT_RATIO
    ):
        raise EVMRevertError("R")

    ratio = uint256(sqrt_price_x96) << 32

    r: int = ratio
    msb: int = 0

    f = yul.shl(7, yul.gt(r, 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF))
    msb = yul.or_(msb, f)
    r = yul.shr(f, r)

    f = yul.shl(7, yul.gt(r, 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF))
    msb = yul.or_(msb, f)
    r = yul.shr(f, r)

    f = yul.shl(6, yul.gt(r, 0xFFFFFFFFFFFFFFFF))
    msb = yul.or_(msb, f)
    r = yul.shr(f, r)

    f = yul.shl(5, yul.gt(r, 0xFFFFFFFF))
    msb = yul.or_(msb, f)
    r = yul.shr(f, r)

    f = yul.shl(4, yul.gt(r, 0xFFFF))
    msb = yul.or_(msb, f)
    r = yul.shr(f, r)

    f = yul.shl(3, yul.gt(r, 0xFF))
    msb = yul.or_(msb, f)
    r = yul.shr(f, r)

    f = yul.shl(2, yul.gt(r, 0xF))
    msb = yul.or_(msb, f)
    r = yul.shr(f, r)

    f = yul.shl(1, yul.gt(r, 0x3))
    msb = yul.or_(msb, f)
    r = yul.shr(f, r)

    f = yul.gt(r, 0x1)
    msb = yul.or_(msb, f)

    if msb >= 128:
        r = ratio >> (msb - 127)
    else:
        r = ratio << (127 - msb)

    log_2 = (int(msb) - 128) << 64

    r = yul.shr(127, yul.mul(r, r))
    f = yul.shr(128, r)
    log_2 = yul.or_(log_2, yul.shl(63, f))
    r = yul.shr(f, r)

    r = yul.shr(127, yul.mul(r, r))
    f = yul.shr(128, r)
    log_2 = yul.or_(log_2, yul.shl(62, f))
    r = yul.shr(f, r)

    r = yul.shr(127, yul.mul(r, r))
    f = yul.shr(128, r)
    log_2 = yul.or_(log_2, yul.shl(61, f))
    r = yul.shr(f, r)

    r = yul.shr(127, yul.mul(r, r))
    f = yul.shr(128, r)
    log_2 = yul.or_(log_2, yul.shl(60, f))
    r = yul.shr(f, r)

    r = yul.shr(127, yul.mul(r, r))
    f = yul.shr(128, r)
    log_2 = yul.or_(log_2, yul.shl(59, f))
    r = yul.shr(f, r)

    r = yul.shr(127, yul.mul(r, r))
    f = yul.shr(128, r)
    log_2 = yul.or_(log_2, yul.shl(58, f))
    r = yul.shr(f, r)

    r = yul.shr(127, yul.mul(r, r))
    f = yul.shr(128, r)
    log_2 = yul.or_(log_2, yul.shl(57, f))
    r = yul.shr(f, r)

    r = yul.shr(127, yul.mul(r, r))
    f = yul.shr(128, r)
    log_2 = yul.or_(log_2, yul.shl(56, f))
    r = yul.shr(f, r)

    r = yul.shr(127, yul.mul(r, r))
    f = yul.shr(128, r)
    log_2 = yul.or_(log_2, yul.shl(55, f))
    r = yul.shr(f, r)

    r = yul.shr(127, yul.mul(r, r))
    f = yul.shr(128, r)
    log_2 = yul.or_(log_2, yul.shl(54, f))
    r = yul.shr(f, r)

    r = yul.shr(127, yul.mul(r, r))
    f = yul.shr(128, r)
    log_2 = yul.or_(log_2, yul.shl(53, f))
    r = yul.shr(f, r)

    r = yul.shr(127, yul.mul(r, r))
    f = yul.shr(128, r)
    log_2 = yul.or_(log_2, yul.shl(52, f))
    r = yul.shr(f, r)

    r = yul.shr(127, yul.mul(r, r))
    f = yul.shr(128, r)
    log_2 = yul.or_(log_2, yul.shl(51, f))
    r = yul.shr(f, r)

    r = yul.shr(127, yul.mul(r, r))
    f = yul.shr(128, r)
    log_2 = yul.or_(log_2, yul.shl(50, f))

    log_sqrt10001 = log_2 * 255738958999603826347141  # 128.128 number

    tick_low = int24(
        (log_sqrt10001 - 3402992956809132418596140100660247210) >> 128
    )
    tick_high = int24(
        (log_sqrt10001 + 291339464771989622907027621153398088495) >> 128
    )

    tick = (
        tick_low
        if (tick_low == tick_high)
        else (
            tick_high
            if getSqrtRatioAtTick(tick_high) <= sqrt_price_x96
            else tick_low
        )
    )

    return tick



from itertools import cycle
from typing import Iterable, List, Union

import eth_abi
from eth_typing import ChecksumAddress
from eth_utils import to_checksum_address
from web3 import Web3


def decode_v3_path(path: bytes) -> List[Union[ChecksumAddress, int]]:
    """
    Decode the `path` byte string used by the Uniswap V3 Router/Router2 contracts.
    `path` is a close-packed encoding of pool addresses and fees.
    """

    path_pos = 0
    decoded_path: List[Union[ChecksumAddress, int]] = []

    # read alternating 20 and 3 byte chunks from the encoded path,
    # store each address (hex string) and fee (int)
    for byte_length in cycle((20, 3)):
        if byte_length == 20:
            address = to_checksum_address(
                path[path_pos : path_pos + byte_length].hex()
            )
            decoded_path.append(address)
        elif byte_length == 3:
            fee = int(
                path[path_pos : path_pos + byte_length].hex(),
                16,
            )
            decoded_path.append(fee)

        path_pos += byte_length

        if path_pos == len(path):
            break

    return decoded_path


def generate_v3_pool_address(
    token_addresses: Iterable[str],
    fee: int,
    factory_address: str,
    init_hash: str,
) -> ChecksumAddress:
    """
    Generate the deterministic pool address from the token addresses and fee.

    Adapted from https://github.com/Uniswap/v3-periphery/blob/main/contracts/libraries/PoolAddress.sol
    """

    token_addresses = sorted([address.lower() for address in token_addresses])

    return to_checksum_address(
        Web3.keccak(
            hexstr="0xff"
            + factory_address[2:]
            + Web3.keccak(
                eth_abi.encode(
                    types=("address", "address", "uint24"),
                    args=(*token_addresses, fee),
                )
            ).hex()[2:]
            + init_hash[2:]
        )[12:]
    )



from abc import ABC
from eth_typing import ChecksumAddress


class ArbitrageHelper(ABC):
    pass


class HelperManager(ABC):
    """
    An abstract base class for managers that generate, track and distribute various helper classes
    """

    pass


class PoolHelper(ABC):
    address: ChecksumAddress


class TokenHelper(ABC):
    pass


class TransactionHelper(ABC):
    pass



from typing import Optional, Union

from eth_typing import ChecksumAddress
from eth_utils import to_checksum_address

from ..config import get_web3
from .abi import UNISWAP_V3_TICKLENS_ABI


class TickLens:
    def __init__(
        self,
        address: Union[str, ChecksumAddress],
        abi: Optional[list] = None,
    ):
        _web3 = get_web3()
        if _web3 is not None:
            self._w3 = _web3
        else:
            from brownie import web3 as brownie_web3  # type: ignore[import]

            if brownie_web3.isConnected():
                self._w3 = brownie_web3
            else:
                raise ValueError("No connected web3 object provided.")

        self.address = to_checksum_address(address)

        self._w3_contract = self._w3.eth.contract(
            address=self.address,
            abi=abi or UNISWAP_V3_TICKLENS_ABI,
        )



from typing import List, Optional, Tuple
from scipy.optimize import minimize_scalar  # type: ignore[import]

from ..erc20_token import Erc20Token
from ..baseclasses import ArbitrageHelper
from ..uniswap.v2_liquidity_pool import LiquidityPool


class FlashBorrowToLpSwapWithFuture(ArbitrageHelper):
    def __init__(
        self,
        borrow_pool: LiquidityPool,
        borrow_token: Erc20Token,
        repay_token: Erc20Token,
        swap_pool_addresses: Optional[List[str]] = None,
        swap_pools: Optional[List[LiquidityPool]] = None,
        name: str = "",
        update_method="polling",
    ):
        if swap_pools is None:
            swap_pools = []

        if swap_pool_addresses is None:
            swap_pool_addresses = []

        if not (swap_pools or swap_pool_addresses):
            raise ValueError(
                "At least one pool address or LiquidityPool object must be provided"
            )

        if swap_pool_addresses and swap_pools:
            raise ValueError(
                "Choose pool addresses or LiquidityPool objects, not both"
            )

        if update_method not in [
            "polling",
            "external",
        ]:
            raise ValueError("update_method must be 'polling' or 'external'")

        if update_method == "external" and swap_pool_addresses:
            raise ValueError(
                "swap pools by address must be updated with the 'polling' method"
            )

        self.borrow_pool = borrow_pool
        self.borrow_token = borrow_token
        self.repay_token = repay_token
        self._update_method = update_method

        # if the object was initialized with pool objects directly, use these directly
        if swap_pools:
            self.swap_pools = swap_pools

        # otherwise, create internal objects
        else:
            self.swap_pools = []
            for address in swap_pool_addresses:
                self.swap_pools.append(
                    LiquidityPool(
                        address=address,
                        update_method=update_method,
                    )
                )

        self.swap_pool_addresses = [pool.address for pool in self.swap_pools]
        self.swap_pool_tokens = [
            [pool.token0, pool.token1] for pool in self.swap_pools
        ]

        self.all_pool_addresses = [
            pool.address for pool in self.swap_pools + [self.borrow_pool]
        ]

        if name:
            self.name = name
        else:
            self.name = (
                self.borrow_pool.name
                + " -> "
                + " -> ".join([pool.name for pool in self.swap_pools])
            )

        if self.borrow_token.address not in [
            self.swap_pools[0].token0.address,
            self.swap_pools[0].token1.address,
        ]:
            raise ValueError("Borrowed token not found in the first swap pool")

        if self.repay_token.address not in [
            self.swap_pools[-1].token0.address,
            self.swap_pools[-1].token1.address,
        ]:
            raise ValueError("Repay token not found in the last swap pool")

        if self.swap_pools[0].token0.address == borrow_token.address:
            forward_token_address = self.swap_pools[0].token1.address
        else:
            forward_token_address = self.swap_pools[0].token0.address

        token_in_address = forward_token_address
        for pool in self.swap_pools:
            if pool.token0.address == token_in_address:
                forward_token_address = pool.token1.address
            elif pool.token1.address == token_in_address:
                forward_token_address = pool.token0.address
            else:
                raise Exception(
                    "Swap pools are invalid, no swap route possible!"
                )

        self.best = {
            "init": True,
            "strategy": "flash borrow swap",
            "borrow_amount": 0,
            "borrow_token": self.borrow_token,
            "borrow_pool": self.borrow_pool,
            "borrow_pool_amounts": [],
            "repay_amount": 0,
            "profit_amount": 0,
            "profit_token": self.repay_token,
            "swap_pools": self.swap_pools,
            "swap_pool_addresses": self.swap_pool_addresses,
            "swap_pool_amounts": [],
            "swap_pool_tokens": self.swap_pool_tokens,
        }

        self.best_future = {
            "strategy": "flash borrow swap",
            "borrow_amount": 0,
            "borrow_token": self.borrow_token,
            "borrow_pool": self.borrow_pool,
            "borrow_pool_amounts": [],
            "repay_amount": 0,
            "profit_amount": 0,
            "profit_token": self.repay_token,
            "swap_pools": self.swap_pools,
            "swap_pool_addresses": self.swap_pool_addresses,
            "swap_pool_amounts": [],
            "swap_pool_tokens": self.swap_pool_tokens,
        }

        # bugfix: maintain a record of the reserve state for associated LPs, used to avoid arb recalculation
        self.reserves: dict = {}
        if self._update_method != "external":
            self.update_reserves()

        # track the gas estimate to execute this arb
        self.gas_estimate = 0

    def _build_multipool_amounts_out(
        self,
        token_in: Erc20Token,
        token_in_quantity: int,
        pool_overrides: Optional[
            List[Tuple[LiquidityPool, Tuple[int, int]]]
        ] = None,
    ) -> List[List[int]]:
        number_of_pools = len(self.swap_pools)

        if pool_overrides is None:
            pool_overrides = []

        pools_amounts_out = []

        for i in range(number_of_pools):
            # determine the output token for pool0
            if token_in.address == self.swap_pools[i].token0.address:
                token_out = self.swap_pools[i].token1
            elif token_in.address == self.swap_pools[i].token1.address:
                token_out = self.swap_pools[i].token0
            else:
                print("wtf?")
                raise Exception

            # value of 0 will result in the default behavior (no reserve overrides)
            override_reserves_token0 = 0
            override_reserves_token1 = 0

            # override the reserves if found in pool_overrides
            for override in pool_overrides:
                if override[0] == self.swap_pools[i]:
                    (
                        override_reserves_token0,
                        override_reserves_token1,
                    ) = override[1]

            # calculate the swap output through pool[i]
            token_out_quantity = self.swap_pools[
                i
            ].calculate_tokens_out_from_tokens_in(
                token_in=token_in,
                token_in_quantity=token_in_quantity,
                override_reserves_token0=override_reserves_token0,
                override_reserves_token1=override_reserves_token1,
            )

            if token_in.address == self.swap_pools[i].token0.address:
                pools_amounts_out.append([0, token_out_quantity])
            elif token_in.address == self.swap_pools[i].token1.address:
                pools_amounts_out.append([token_out_quantity, 0])

            if i == number_of_pools - 1:
                break
            else:
                # otherwise, feed the results back into the loop
                token_in = token_out
                token_in_quantity = token_out_quantity

        return pools_amounts_out

    def _calculate_arbitrage(
        self,
        override_future: bool = False,
        pool_overrides: Optional[
            List[Tuple[LiquidityPool, Tuple[int, int]]]
        ] = None,
    ):
        if pool_overrides is None:
            pool_overrides = []

        borrow_pool_reserves_token0 = self.borrow_pool.reserves_token0
        borrow_pool_reserves_token1 = self.borrow_pool.reserves_token1

        # override the borrowing pool reserves if any of the pools in pool_overrides
        # reference this pool
        if override_future:
            for override in pool_overrides:
                if override[0] == self.borrow_pool:
                    borrow_pool_reserves_token0 = override[1][0]
                    borrow_pool_reserves_token1 = override[1][1]
                    break

        # set up the boundaries for the Brent optimizer based on which token
        # is being borrowed
        if self.borrow_token.address == self.borrow_pool.token0.address:
            bounds = (
                0,
                float(borrow_pool_reserves_token0),
            )
            bracket = (
                0.001 * borrow_pool_reserves_token0,
                0.01 * borrow_pool_reserves_token0,
            )
        elif self.borrow_token.address == self.borrow_pool.token1.address:
            bounds = (
                0,
                float(borrow_pool_reserves_token1),
            )
            bracket = (
                0.001 * borrow_pool_reserves_token1,
                0.01 * borrow_pool_reserves_token1,
            )
        else:
            print("_calculate_arbitrage: WTF? Could not identify borrow token")
            raise Exception

        try:
            opt = minimize_scalar(
                lambda x: -float(
                    self.calculate_multipool_tokens_out_from_tokens_in(
                        token_in=self.borrow_token,
                        token_in_quantity=x,
                        pool_overrides=pool_overrides,
                    )
                    - self.borrow_pool.calculate_tokens_in_from_tokens_out(
                        token_in=self.repay_token,
                        token_out_quantity=x,
                        override_reserves_token0=borrow_pool_reserves_token0,
                        override_reserves_token1=borrow_pool_reserves_token1,
                    )
                ),
                method="bounded",
                bounds=bounds,
                bracket=bracket,
                options={
                    "xatol": 1.0,
                    # "disp": 3,
                },
            )
        except Exception as e:
            print(e)
            print(f"bounds: {bounds}")
            print(f"bracket: {bracket}")
            raise
        else:
            best_borrow = int(opt.x)

        if self.borrow_token.address == self.borrow_pool.token0.address:
            borrow_amounts = [best_borrow, 0]
        elif self.borrow_token.address == self.borrow_pool.token1.address:
            borrow_amounts = [0, best_borrow]
        else:
            print("_calculate_arbitrage: WTF? Could not identify borrow token")
            raise Exception

        best_repay = self.borrow_pool.calculate_tokens_in_from_tokens_out(
            token_in=self.repay_token,
            token_out_quantity=best_borrow,
            override_reserves_token0=borrow_pool_reserves_token0,
            override_reserves_token1=borrow_pool_reserves_token1,
        )
        best_profit = -int(opt.fun)

        if override_future:
            if best_borrow > 0 and best_profit > 0:
                self.best_future.update(
                    {
                        "borrow_amount": best_borrow,
                        "borrow_pool_amounts": borrow_amounts,
                        "repay_amount": best_repay,
                        "profit_amount": best_profit,
                        "swap_pool_amounts": self._build_multipool_amounts_out(
                            token_in=self.borrow_token,
                            token_in_quantity=best_borrow,
                            pool_overrides=pool_overrides,
                        ),
                    }
                )
            else:
                self.best_future.update(
                    {
                        "borrow_amount": 0,
                        "borrow_pool_amounts": [],
                        "repay_amount": 0,
                        "profit_amount": 0,
                        "swap_pool_amounts": [],
                    }
                )
        else:
            # only save opportunities with rational, positive values
            if best_borrow > 0 and best_profit > 0:
                self.best.update(
                    {
                        "borrow_amount": best_borrow,
                        "borrow_pool_amounts": borrow_amounts,
                        "repay_amount": best_repay,
                        "profit_amount": best_profit,
                        "swap_pool_amounts": self._build_multipool_amounts_out(
                            token_in=self.borrow_token,
                            token_in_quantity=best_borrow,
                        ),
                    }
                )
            else:
                self.best.update(
                    {
                        "borrow_amount": 0,
                        "borrow_pool_amounts": [],
                        "repay_amount": 0,
                        "profit_amount": 0,
                        "swap_pool_amounts": [],
                    }
                )

    def __str__(self) -> str:
        return self.name

    def calculate_multipool_tokens_out_from_tokens_in(
        self,
        token_in: Erc20Token,
        token_in_quantity: int,
        pool_overrides: Optional[
            List[Tuple[LiquidityPool, Tuple[int, int]]]
        ] = None,
    ) -> int:
        """
        Calculates the expected token OUTPUT from the last pool for a given token INPUT to the first pool
        at current pool reserves. Uses the self.token0 and self.token1 pointers to determine which token
        is being swapped in and uses the appropriate formula

        2022-06-14 update: add support for overriding pool reserves
        """

        if pool_overrides is None:
            pool_overrides = []

        number_of_pools = len(self.swap_pools)

        for i in range(number_of_pools):
            # determine the output token for pool0
            if token_in.address == self.swap_pools[i].token0.address:
                token_out = self.swap_pools[i].token1
            elif token_in.address == self.swap_pools[i].token1.address:
                token_out = self.swap_pools[i].token0
            else:
                print("wtf?")
                raise Exception

            override_reserves_token0 = 0
            override_reserves_token1 = 0

            for override_pool, override_reserves in pool_overrides:
                # each override is a tuple of form (LiquidityPool, (reserve0,reserve1))
                if override_pool == self.swap_pools[i]:
                    override_reserves_token0 = override_reserves[0]
                    override_reserves_token1 = override_reserves[1]
                    break

            # calculate the swap output through pool[i]
            token_out_quantity = self.swap_pools[
                i
            ].calculate_tokens_out_from_tokens_in(
                token_in=token_in,
                token_in_quantity=token_in_quantity,
                override_reserves_token0=override_reserves_token0,
                override_reserves_token1=override_reserves_token1,
            )

            if i == number_of_pools - 1:
                break
            else:
                # otherwise, use the output as input on the next loop
                token_in = token_out
                token_in_quantity = token_out_quantity

        return token_out_quantity

    def clear_best(self):
        self.best.update(
            {
                "borrow_amount": 0,
                "borrow_pool_amounts": [],
                "repay_amount": 0,
                "profit_amount": 0,
                "swap_pool_amounts": [],
            }
        )

    def clear_best_future(self):
        self.best_future.update(
            {
                "borrow_amount": 0,
                "borrow_pool_amounts": [],
                "repay_amount": 0,
                "profit_amount": 0,
                "swap_pool_amounts": [],
            }
        )

    def update_reserves(
        self,
        silent: bool = False,
        print_reserves: bool = True,
        print_ratios: bool = True,
        override_future: bool = False,
        pool_overrides: Optional[
            List[Tuple[LiquidityPool, Tuple[int, int]]]
        ] = None,
    ) -> bool:
        """
        Updates reserve values for one or more liquidity pools by calling update_reserves(), which returns False if the reserves have not changed.
        Will calculate arbitrage amounts only after checking all pools and finding a reason to update, or on startup (via the 'init' dictionary key)
        """

        if pool_overrides is None:
            pool_overrides = []

        recalculate = False

        if self._update_method != "external":
            # calculate initial arbitrage after the object is instantiated, otherwise proceed with normal checks
            if self.best["init"] is True:
                self.best["init"] = False
                recalculate = True

            # flag for recalculation if the borrowing pool has been updated
            if self.borrow_pool.update_reserves(
                silent=silent,
                print_reserves=print_reserves,
                print_ratios=print_ratios,
            ):
                recalculate = True

            # flag for recalculation if any of the pools along the swap path have been updated
            for pool in self.swap_pools:
                if pool.update_reserves(
                    silent=silent,
                    print_reserves=print_reserves,
                    print_ratios=print_ratios,
                ):
                    recalculate = True

        if override_future:
            if not pool_overrides:
                raise ValueError("Overrides must be provided!")

            recalculate = True

            for override_pool, override_reserves in pool_overrides:
                if type(override_pool) is not LiquidityPool:
                    raise TypeError(
                        "Override does not include a LiquidityPool object!"
                    )

                if type(override_reserves) is not tuple:
                    raise TypeError("Overrides not formatted as a tuple")

                if len(override_reserves) != 2:
                    raise ValueError("Override length must be 2")

                if not isinstance(override_reserves[0], int):
                    raise TypeError(
                        f"override for token0 must be int, is {type(override_reserves[0])}"
                    )

                if not isinstance(override_reserves[1], int):
                    raise TypeError(
                        f"override for token1 must be int, is {type(override_reserves[1])}"
                    )

        # update the reserves tracked in self.reserves and flag the arb for recalculation if they did not match
        for pool in [self.borrow_pool] + self.swap_pools:
            current_lp_reserves = (pool.reserves_token0, pool.reserves_token1)
            if self.reserves.get(pool.address) != current_lp_reserves:
                self.reserves[pool.address] = current_lp_reserves
                recalculate = True

        if recalculate:
            self._calculate_arbitrage(
                override_future=override_future,
                pool_overrides=pool_overrides,
            )
            return True
        else:
            return False



from eth_utils import to_checksum_address
from degenbot.erc20_token import Erc20Token


class MockErc20Token(Erc20Token):
    def __init__(self):
        pass


def test_erc20token_comparisons():
    token0 = MockErc20Token()
    token0.address = to_checksum_address(
        "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"
    )

    token1 = MockErc20Token()
    token1.address = to_checksum_address(
        "0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599"
    )

    assert token0 == "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"
    assert token0 == "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2".lower()
    assert token0 == to_checksum_address(
        "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"
    )

    assert token0 != token1

    assert token0 > token1
    assert token0 > "0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599"
    assert token0 > "0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599".lower()
    assert token0 > to_checksum_address(
        "0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599"
    )

    assert token1 < token0
    assert token1 < "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"
    assert token1 < "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2".lower()
    assert token1 < to_checksum_address(
        "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"
    )



def gt(x, y):
    return 1 if x > y else 0


def lt(x, y):
    return 1 if x < y else 0


def mod(x, y):
    return 0 if y == 0 else x % y


def mul(x, y):
    return x * y


def mulmod(x, y, m):
    return 0 if m == 0 else (x * y) % m


def shl(x, y):
    return y << x


def shr(x, y):
    return y >> x


def or_(x, y):
    return x | y


def not_(x):
    return ~x


def add(x, y):
    return x + y


def sub(x, y):
    return x - y


def div(x, y):
    return 0 if y == 0 else x // y



import dataclasses
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .v2_liquidity_pool import LiquidityPool


@dataclasses.dataclass(slots=True, frozen=True)
class UniswapV2PoolState:
    pool: "LiquidityPool"
    reserves_token0: int
    reserves_token1: int


@dataclasses.dataclass(slots=True, frozen=True)
class UniswapV2PoolSimulationResult:
    amount0_delta: int
    amount1_delta: int
    current_state: UniswapV2PoolState
    future_state: UniswapV2PoolState
